@@ -1,16 +1,11 @@
 from __future__ import annotations
 
-from collections.abc import Mapping, Sequence
-from typing import TYPE_CHECKING, Any, Union
+from collections.abc import Sequence
+from typing import TYPE_CHECKING
 from urllib.parse import unquote, urldefrag, urljoin
 
-from pyrsistent import m, s
-from pyrsistent.typing import PMap, PSet
-
-try:
-    Mapping[str, str]
-except TypeError:
-    from typing import Mapping
+from pyrsistent import m, plist, s
+from pyrsistent.typing import PList, PMap, PSet
 
 
 class UnsupportedSubclassing(Exception):
@@ -29,6 +24,8 @@ class UnidentifiedResource(Exception):
 
 if TYPE_CHECKING:
     from attrs import define, evolve, field, frozen
+
+    from referencing.typing import AnchorType, Schema, Specification
 else:
     from attrs import define as _define, evolve, field, frozen as _frozen
 
@@ -41,9 +38,6 @@ else:
         return _frozen(cls)
 
 
-Schema = Union[bool, Mapping[str, Any]]
-
-
 @frozen
 class Anchor:
 
@@ -51,6 +45,9 @@ class Anchor:
     name: str
     resource: Schema
 
+    def resolve(self, dynamic_scope, uri) -> tuple[Schema, str]:
+        return self.resource, uri
+
 
 @frozen
 class DynamicAnchor:
@@ -59,8 +56,26 @@ class DynamicAnchor:
     name: str
     resource: Schema
 
+    def resolve(self, dynamic_scope, uri) -> tuple[Schema, str]:
+        last = self.resource
+        for resource, anchors in dynamic_scope:
+            anchor = anchors.get(self.name)
+            if isinstance(anchor, DynamicAnchor):
+                last = anchor.resource
+            elif "$ref" not in resource:
+                break
+        return last, id_of(last) or ""  # FIXME: consider when this can be None
 
-AnchorType = Union[Anchor, DynamicAnchor]
+
+class OpaqueSpecification:
+    """
+    A non-specification `Specification` which treats resources opaquely.
+
+    In particular, they have no subresources.
+    """
+
+    def subresources_of(self, resource: Schema):
+        return ()
 
 
 @frozen
@@ -70,7 +85,8 @@ class Registry:
         default=m(),
         repr=lambda value: f"({len(value)} entries)",
     )
-    _uncrawled: PSet[str] = field(default=s())
+    _uncrawled: PSet[str] = s()
+    _specification: Specification = OpaqueSpecification()
 
     def update(self, *registries: Registry) -> Registry:
         contents = (each._contents for each in registries)
@@ -121,8 +137,8 @@ class Registry:
             registry = self._crawl()
         return registry._contents[uri][0], registry
 
-    def anchor_at(self, uri, name) -> AnchorType:
-        return self._contents[uri][1][name]
+    def anchors_at(self, uri) -> PMap[str, AnchorType]:
+        return self._contents[uri][1]
 
     def _crawl(self) -> Registry:
         registry = self
@@ -154,27 +170,17 @@ class Registry:
                         resource=resource,
                     ),
                 )
-
-            resources.extend(  # TODO: delay finding anchors in subresources...
-                (uri, resource[k]) for k in SUBRESOURCE if k in resource
-            )
             resources.extend(
-                (uri, subresource)
-                for k in SUBRESOURCE_VALUES
-                if k in resource
-                for subresource in resource[k].values()
-            )
-            resources.extend(
-                (uri, subresource)
-                for k in SUBRESOURCE_ITEMS
-                if k in resource
-                for subresource in resource[k]
+                (uri, each)
+                for each in self._specification.subresources_of(resource)
+                if each is not True and each is not False
             )
         return evolve(registry, uncrawled=s())
 
-    def resolver(self, root) -> Resolver:
+    def resolver(self, root, specification) -> Resolver:
         uri = id_of(root) or ""
         registry = self.with_identified_resource(uri=uri, resource=root)
+        registry = evolve(registry, specification=specification)
         return Resolver(base_uri=uri, registry=registry)
 
 
@@ -183,6 +189,7 @@ class Resolver:
 
     _base_uri: str
     _registry: Registry
+    _previous: PList[Resolver] = plist()
 
     def lookup(self, ref: str) -> tuple[Schema, Resolver]:
         if ref.startswith("#"):
@@ -199,9 +206,13 @@ class Resolver:
                     segment = segment.replace("~1", "/").replace("~0", "~")
                 target = target[segment]  # type: ignore # this can't be a bool
         elif fragment:
-            target = registry.anchor_at(uri=uri, name=fragment).resource
+            anchor = registry.anchors_at(uri=uri)[fragment]
+            target, uri = anchor.resolve(
+                dynamic_scope=self.dynamic_scope(),
+                uri=uri,
+            )
 
-        return target, evolve(self, base_uri=uri, registry=registry)
+        return target, self.evolve(base_uri=uri, registry=registry)
 
     def with_root(self, root) -> Resolver:
         maybe_relative = id_of(root)
@@ -213,12 +224,16 @@ class Resolver:
             uri=uri,
             resource=root,
         )
-        return evolve(self, base_uri=uri, registry=registry)
+        return self.evolve(base_uri=uri, registry=registry)
 
+    def evolve(self, **kwargs):
+        previous = self._previous.cons(self._base_uri)
+        return evolve(self, previous=previous, **kwargs)
 
-SUBRESOURCE = {"items", "not"}
-SUBRESOURCE_ITEMS = {"allOf"}
-SUBRESOURCE_VALUES = {"$defs", "properties"}
+    def dynamic_scope(self):
+        for uri in self._previous:
+            resource, _ = self._registry.resource_at(uri)
+            yield resource, self._registry.anchors_at(uri)
 
 
 def id_of(resource) -> str | None:
